import * as TE from "fp-ts/lib/TaskEither";
import * as O from "fp-ts/lib/Option";
import * as RA from "fp-ts/ReadonlyArray";
import { pipe } from "fp-ts/lib/function";
import { readConfig } from "./config";
import { parseArgs, type Args } from "./args";
import { executorFor, type Test, type TestType } from "./canary";
import { constructJobsTable, perform, type Job } from "./canary/scheduler";
import { publish, PublisherType } from "./publisher";
import { ConsoleLogger } from "./util";

const testFilters = (args: Args): ReadonlyArray<(test: Test) => boolean> =>
  (
    [
      [args.testType, (testType: TestType) => (test: Test) => test.type === testType],
      [args.testName, (testName: string) => (test: Test) => testName === test.name]
    ] as ReadonlyArray<[O.Option<any>, (t: any) => (test: Test) => boolean]>
  )
    .map(([o, pred]) =>
      pipe(
        o,
        O.map((val) => pred(val))
      )
    )
    .filter(O.isSome)
    .map(({ value }) => value);

const main: TE.TaskEither<string | Error, any> = pipe(
  TE.fromEither(parseArgs(Bun.argv)),
  TE.bindTo("args"),
  TE.bind("now", () => TE.fromIO(() => new Date())),
  TE.bind("config", ({ args }) => TE.fromEither(readConfig(args.testsFile))),
  TE.flatMap(({ config, args, now }) => {
    const filters = testFilters(args);
    const jobs = config.tests
      .filter((test) => filters.every((filter) => filter(test)))
      .map((test, i): Job => {
        const toString = () => test.name;
        const id = i.toString();
        const schedule = test.schedule;
        const maxRetries = 3;
        const execute = executorFor(test.type, test.job);
        return { toString, id, schedule, maxRetries, execute };
      });
    const jobsTable = constructJobsTable(jobs, now);
    const publishers = pipe([{ publisherType: PublisherType.CONSOLE, post: ConsoleLogger }] as const, RA.map(publish));

    return pipe(perform(jobsTable, jobs, publishers, new Date()), TE.fromTask);
  })
);
main().then(() => process.exit());
